package ru.dbotthepony.mc.otm.registry

import net.minecraft.advancements.CriteriaTriggers
import net.minecraft.client.renderer.item.ItemProperties
import net.minecraft.core.BlockPos
import net.minecraft.resources.ResourceLocation
import net.minecraft.world.damagesource.DamageSource
import net.minecraft.world.entity.EntityType
import net.minecraft.world.level.BlockGetter
import net.minecraft.world.level.block.*
import net.minecraft.world.level.block.state.BlockBehaviour
import net.minecraft.world.level.block.state.BlockState
import net.minecraft.world.level.material.Material
import net.minecraft.world.level.material.MaterialColor
import net.minecraftforge.eventbus.api.IEventBus
import net.minecraftforge.fml.event.lifecycle.FMLClientSetupEvent
import net.minecraftforge.fml.event.lifecycle.FMLCommonSetupEvent
import net.minecraftforge.registries.NewRegistryEvent
import ru.dbotthepony.mc.otm.OverdriveThatMatters
import ru.dbotthepony.mc.otm.android.AndroidFeatureType
import ru.dbotthepony.mc.otm.android.feature.EnderTeleporterFeature
import ru.dbotthepony.mc.otm.android.feature.FallDampenersFeature
import ru.dbotthepony.mc.otm.android.feature.ItemMagnetFeature
import ru.dbotthepony.mc.otm.android.feature.JumpBoostFeature
import ru.dbotthepony.mc.otm.android.feature.NanobotsArmorFeature
import ru.dbotthepony.mc.otm.android.feature.ShockwaveFeature
import ru.dbotthepony.mc.otm.block.decorative.CargoCrateBlock
import ru.dbotthepony.mc.otm.block.decorative.TritaniumPressurePlate
import ru.dbotthepony.mc.otm.capability.matteryEnergy
import ru.dbotthepony.mc.otm.core.math.Decimal
import ru.dbotthepony.mc.otm.item.EnergySwordItem
import ru.dbotthepony.mc.otm.registry.objects.ColoredDecorativeBlock
import ru.dbotthepony.mc.otm.registry.objects.CrateProperties
import ru.dbotthepony.mc.otm.registry.objects.DecorativeBlock
import ru.dbotthepony.mc.otm.registry.objects.StripedColoredDecorativeBlock
import ru.dbotthepony.mc.otm.triggers.AndroidBatteryTrigger
import ru.dbotthepony.mc.otm.triggers.KillAsAndroidTrigger
import ru.dbotthepony.mc.otm.triggers.AndroidResearchTrigger
import ru.dbotthepony.mc.otm.triggers.AndroidTravelUnderwater
import ru.dbotthepony.mc.otm.triggers.BecomeAndroidDeathTrigger
import ru.dbotthepony.mc.otm.triggers.BecomeAndroidSleepTrigger
import ru.dbotthepony.mc.otm.triggers.BecomeAndroidTrigger
import ru.dbotthepony.mc.otm.triggers.BecomeHumaneTrigger
import ru.dbotthepony.mc.otm.triggers.BlackHoleTrigger
import ru.dbotthepony.mc.otm.triggers.EnderTeleporterFallDeathTrigger
import ru.dbotthepony.mc.otm.triggers.FallDampenersSaveTrigger
import ru.dbotthepony.mc.otm.triggers.NanobotsArmorTrigger
import ru.dbotthepony.mc.otm.triggers.PhantomSpawnDeniedTrigger
import ru.dbotthepony.mc.otm.triggers.ShockwaveDamageMobTrigger
import ru.dbotthepony.mc.otm.triggers.ShockwaveTrigger

object MRegistry {
	private val features = RegistryDelegate<AndroidFeatureType<*>>("android_features")
	val ANDROID_FEATURES by features
	val ANDROID_FEATURES_LOCATION get() = features.location
	val ANDROID_FEATURES_KEY get() = features.key

	private fun register(event: NewRegistryEvent) {
		features.build(event)
	}

	val CARGO_CRATES = DecorativeBlock(MNames.CARGO_CRATE, ::CargoCrateBlock)

	val DECORATIVE_CRATE = DecorativeBlock.simple(MNames.DECORATIVE_CRATE) {
		BlockBehaviour.Properties.of(Material.METAL, it?.materialColor ?: MaterialColor.SNOW)
			.sound(SoundType.METAL)
			.requiresCorrectToolForDrops()
			.explosionResistance(10f)
			.destroyTime(1f)
	}

	val TRITANIUM_BLOCK = DecorativeBlock.simple(MNames.TRITANIUM_BLOCK) {
		BlockBehaviour.Properties.of(Material.METAL, it?.materialColor ?: MaterialColor.COLOR_LIGHT_BLUE)
			.sound(SoundType.BASALT)
			.requiresCorrectToolForDrops()
			.explosionResistance(80f)
			.destroyTime(2.5f)
	}

	val TRITANIUM_STAIRS = DecorativeBlock(MNames.TRITANIUM_STAIRS) {
		StairBlock(
			{ TRITANIUM_BLOCK.allBlocks[it]!!.defaultBlockState() },
			BlockBehaviour.Properties.copy(TRITANIUM_BLOCK.allBlocks[it]!!)
		)
	}

	val TRITANIUM_SLAB = DecorativeBlock(MNames.TRITANIUM_SLAB) {
		SlabBlock(BlockBehaviour.Properties.copy(TRITANIUM_BLOCK.allBlocks[it]!!))
	}

	val TRITANIUM_WALL = DecorativeBlock(MNames.TRITANIUM_WALL) {
		WallBlock(BlockBehaviour.Properties.copy(TRITANIUM_BLOCK.allBlocks[it]!!))
	}

	val TRITANIUM_PRESSURE_PLATE = DecorativeBlock(MNames.TRITANIUM_PRESSURE_PLATE, ::TritaniumPressurePlate)

	val VENT = DecorativeBlock.simple(MNames.VENT) {
		BlockBehaviour.Properties.of(Material.METAL, it?.materialColor ?: MaterialColor.COLOR_LIGHT_BLUE)
			.sound(SoundType.BASALT)
			.requiresCorrectToolForDrops()
			.explosionResistance(20f)
			.destroyTime(1.5f)
	}

	val VENT_ALTERNATIVE = DecorativeBlock.simple(MNames.VENT_ALTERNATIVE) {
		BlockBehaviour.Properties.of(Material.METAL, it?.materialColor ?: MaterialColor.COLOR_LIGHT_BLUE)
			.sound(SoundType.BASALT)
			.requiresCorrectToolForDrops()
			.explosionResistance(20f)
			.destroyTime(1.5f)
	}

	val FLOOR_TILES = ColoredDecorativeBlock.simple(MNames.FLOOR_TILES) {
		BlockBehaviour.Properties.of(Material.STONE, it.materialColor)
			.sound(SoundType.STONE)
			.requiresCorrectToolForDrops()
			.strength(1.5f, 6f)
	}

	val FLOOR_TILES_STAIRS = ColoredDecorativeBlock(MNames.FLOOR_TILES_STAIRS) {
		StairBlock(
			{ FLOOR_TILES.blocks[it]!!.defaultBlockState() },
			BlockBehaviour.Properties.copy(FLOOR_TILES.blocks[it]!!)
		)
	}

	val FLOOR_TILES_SLAB = ColoredDecorativeBlock(MNames.FLOOR_TILES_SLAB) {
		SlabBlock(BlockBehaviour.Properties.copy(FLOOR_TILES.blocks[it]!!))
	}

	val UNREFINED_FLOOR_TILES = ColoredDecorativeBlock.simple(MNames.UNREFINED_FLOOR_TILES) {
		BlockBehaviour.Properties.of(Material.CLAY, it.materialColor)
			.sound(SoundType.GRAVEL)
			.strength(1f, 2f)
	}

	val INDUSTRIAL_GLASS = DecorativeBlock(MNames.INDUSTRIAL_GLASS) { color ->
		val properties =
			BlockBehaviour.Properties.of(Material.GLASS, if (color != null) color.materialColor else MaterialColor.NONE)
				.destroyTime(1.5f)
				.explosionResistance(40f)
				.requiresCorrectToolForDrops()
				.sound(SoundType.GLASS)
				.noOcclusion()
				.isValidSpawn { _: BlockState, _: BlockGetter, _: BlockPos, _: EntityType<*>? -> false }
				.isRedstoneConductor { _: BlockState, _: BlockGetter, _: BlockPos -> false }
				.isSuffocating { _: BlockState, _: BlockGetter, _: BlockPos -> false }
				.isViewBlocking { _: BlockState, _: BlockGetter, _: BlockPos -> false }

		if (color != null) {
			return@DecorativeBlock StainedGlassBlock(color, properties)
		}

		return@DecorativeBlock GlassBlock(properties)
	}

	val INDUSTRIAL_GLASS_PANE = DecorativeBlock(MNames.INDUSTRIAL_GLASS_PANE) { color ->
		val properties =
			BlockBehaviour.Properties.of(Material.GLASS, if (color != null) color.materialColor else MaterialColor.NONE)
				.strength(1.25f, 5.0f)
				.requiresCorrectToolForDrops()
				.sound(SoundType.GLASS)
				.noOcclusion()

		if (color != null) {
			return@DecorativeBlock StainedGlassPaneBlock(color, properties)
		}

		return@DecorativeBlock IronBarsBlock(properties)
	}

	val CRATE_RED = CrateProperties(MaterialColor.COLOR_RED, "crate_red")
	val CRATE_BLUE = CrateProperties(MaterialColor.COLOR_BLUE, "crate_blue")
	val CRATE_YELLOW = CrateProperties(MaterialColor.COLOR_YELLOW, "crate_yellow")
	val CRATE_GREEN = CrateProperties(MaterialColor.COLOR_GREEN, "crate_green")
	val CRATE_BLACK = CrateProperties(MaterialColor.COLOR_BLACK, "crate_black")
	val CRATE_PINK = CrateProperties(MaterialColor.COLOR_PINK, "crate_pink")
	val CRATE_PURPLE = CrateProperties(MaterialColor.COLOR_PURPLE, "crate_purple")

	val CRATE_LIST = listOf(
		CRATE_RED,
		CRATE_BLUE,
		CRATE_YELLOW,
		CRATE_GREEN,
		CRATE_BLACK,
		CRATE_PINK,
		CRATE_PURPLE,
	)

	const val DAMAGE_BECOME_ANDROID_ID = "otm_become_android"
	const val DAMAGE_BECOME_HUMANE_ID = "otm_become_humane"
	const val DAMAGE_EVENT_HORIZON_ID = "otm_event_horizon"
	const val DAMAGE_HAWKING_RADIATION_ID = "otm_hawking_radiation"
	const val DAMAGE_EXOPACK_PROBE_ID = "otm_exopack_probe"

	val DAMAGE_EXOPACK_PROBE = ImmutableDamageSource(DamageSource(DAMAGE_EXOPACK_PROBE_ID).bypassArmor().bypassMagic())

	val DAMAGE_BECOME_ANDROID = ImmutableDamageSource(DamageSource(DAMAGE_BECOME_ANDROID_ID).bypassArmor().bypassInvul().bypassMagic())
	val DAMAGE_BECOME_HUMANE = ImmutableDamageSource(DamageSource(DAMAGE_BECOME_HUMANE_ID).bypassArmor().bypassInvul().bypassMagic())
	val DAMAGE_EVENT_HORIZON = ImmutableDamageSource(DamageSource(DAMAGE_EVENT_HORIZON_ID).bypassMagic().bypassArmor())
	val DAMAGE_HAWKING_RADIATION = ImmutableDamageSource(DamageSource(DAMAGE_HAWKING_RADIATION_ID))
	const val DAMAGE_EMP_NAME = "otm_emp"
	const val DAMAGE_SHOCKWAVE_NAME = "otm_shockwave"
	const val DAMAGE_PLASMA_NAME = "otm_plasma"
	val DAMAGE_EMP: DamageSource = ImmutableDamageSource(EMPDamageSource())

	val TRITANIUM_STRIPED_BLOCK = StripedColoredDecorativeBlock(MNames.TRITANIUM_STRIPED_BLOCK, { colorA, _ ->
		Block(BlockBehaviour.Properties.of(Material.METAL, colorA.materialColor)
			.sound(SoundType.BASALT)
			.requiresCorrectToolForDrops()
			.explosionResistance(80f)
			.strength(4f))
	})

	val TRITANIUM_STRIPED_STAIRS = StripedColoredDecorativeBlock(MNames.TRITANIUM_STRIPED_STAIRS, { colorA, colorB ->
		StairBlock({ TRITANIUM_STRIPED_BLOCK.getBlock(colorA, colorB).defaultBlockState() }, BlockBehaviour.Properties.copy(TRITANIUM_STRIPED_BLOCK.getBlock(colorA, colorB)))
	})

	val TRITANIUM_STRIPED_SLAB = StripedColoredDecorativeBlock(MNames.TRITANIUM_STRIPED_SLAB, { colorA, colorB ->
		SlabBlock(BlockBehaviour.Properties.copy(TRITANIUM_STRIPED_BLOCK.getBlock(colorA, colorB)))
	})

	val TRITANIUM_STRIPED_WALL = StripedColoredDecorativeBlock(MNames.TRITANIUM_STRIPED_WALL, { colorA, colorB ->
		WallBlock(BlockBehaviour.Properties.copy(TRITANIUM_STRIPED_BLOCK.getBlock(colorA, colorB)))
	})

	fun initialize(bus: IEventBus) {
		bus.addListener(this::register)
		bus.addListener(this::initializeClient)
		bus.addListener(this::initializeCommon)
		bus.addListener(MStats::registerVanilla)

		MBlocks.register(bus)
		MBlockEntities.register(bus)
		MEntityTypes.register(bus)
		MMenus.register(bus)
		MItems.register(bus)
		AndroidFeatures.register(bus)
		MSoundEvents.register(bus)
		LootModifiers.register(bus)
		MItemFunctionTypes.register(bus)
		MLootItemConditions.register(bus)

		MRecipes.register(bus)

		// call static constructors
		NanobotsArmorFeature.Companion
		ShockwaveFeature.Companion
		ItemMagnetFeature.Companion
		FallDampenersFeature.Companion
		JumpBoostFeature.Companion
		EnderTeleporterFeature.Companion
	}

	private fun initializeCommon(event: FMLCommonSetupEvent) {
		event.enqueueWork {
			CriteriaTriggers.register(BlackHoleTrigger)
			CriteriaTriggers.register(BecomeAndroidTrigger)
			CriteriaTriggers.register(BecomeAndroidDeathTrigger)
			CriteriaTriggers.register(BecomeAndroidSleepTrigger)
			CriteriaTriggers.register(BecomeHumaneTrigger)
			CriteriaTriggers.register(AndroidResearchTrigger)
			CriteriaTriggers.register(ShockwaveDamageMobTrigger)
			CriteriaTriggers.register(ShockwaveTrigger)
			CriteriaTriggers.register(PhantomSpawnDeniedTrigger)
			CriteriaTriggers.register(AndroidBatteryTrigger)
			CriteriaTriggers.register(NanobotsArmorTrigger)
			CriteriaTriggers.register(FallDampenersSaveTrigger)
			CriteriaTriggers.register(EnderTeleporterFallDeathTrigger)
			CriteriaTriggers.register(KillAsAndroidTrigger)
			CriteriaTriggers.register(AndroidTravelUnderwater)
		}
	}

	private fun initializeClient(event: FMLClientSetupEvent) {
		ItemProperties.register(MItems.TRITANIUM_SHIELD, ResourceLocation(OverdriveThatMatters.MOD_ID, "blocking")) { stack, level, entity, _ ->
			if (entity?.isUsingItem == true && entity.useItem == stack) {
				1f
			} else {
				0f
			}
		}
		ItemProperties.register(MItems.ENERGY_SWORD, ResourceLocation(OverdriveThatMatters.MOD_ID, "is_powered")) { stack, _, _, _ ->
			if ((stack.matteryEnergy?.batteryLevel ?: Decimal.ZERO) >= EnergySwordItem.ENERGY_PER_SWING) {
				1f
			} else {
				0f
			}
		}
	}
}
