package ru.dbotthepony.mc.otm.tests

import it.unimi.dsi.fastutil.io.FastByteArrayInputStream
import it.unimi.dsi.fastutil.io.FastByteArrayOutputStream
import org.junit.jupiter.api.Assertions.assertEquals
import org.junit.jupiter.api.DisplayName
import org.junit.jupiter.api.Test
import ru.dbotthepony.mc.otm.core.math.Decimal
import java.io.ObjectInputStream
import java.io.ObjectOutputStream

object DecimalTests {
	@Test
	@DisplayName("ImpreciseFraction comparison")
	fun comparison() {
		check(Decimal(642, 0.43774) > Decimal(641, 0.43774)) { "must be bigger" }
		check(Decimal(642, 0.43774) > Decimal(-641, 0.43774)) { "must be bigger" }
		check(Decimal(0) == Decimal(0)) { "integer notation" }
		check(Decimal(0.1) > Decimal(0)) { "must be bigger" }
		check(Decimal(-0.1) < Decimal(0)) { "must be lesser" }
		check(Decimal(-1, -0.1) < Decimal(-1)) { "must be lesser" }
		check(Decimal(0.1) == Decimal(0.1)) { "double notation" }
		check(Decimal("0.1") == Decimal("0.1")) { "string notation" }
	}

	@Test
	@DisplayName("ImpreciseFraction mathematical operations")
	fun math() {
		check((Decimal(1) + Decimal(2)) == Decimal(3)) { "1 + 2 != 3" }
		check((Decimal(0, 0.5) + Decimal(0, 0.5)) == Decimal(1)) { "0.5 + 0.5 != 1" }
		check((Decimal(0, 0.5) + Decimal(0, -0.5)) == Decimal(0)) { "0.5 + -0.5 != 1" }
		check((Decimal(0, 0.5) - Decimal(0, -0.5)) == Decimal(1)) { "0.5 - -0.5 != 1" }
		check((Decimal(0, 0.3) - Decimal(1, 0.2)) == Decimal(0, -0.9)) { "0.3 - 1.2 != -0.9 ${Decimal(0, -0.9)} ${Decimal(0, 0.3) - Decimal(1, 0.2)}" }
		check((Decimal(0, 0.3) * Decimal(0, 0.3)) == Decimal(0, 0.09)) { "0.3 * 0.3 != 0.9 ${Decimal(0, 0.3) * Decimal(0, 0.3)}" }
		check((Decimal(2, 0.3) * Decimal(2, 0.3)) == Decimal(5, 0.29)) { "2.3 * 2.3 != 5.29 ${Decimal(2, 0.3) * Decimal(2, 0.3)}" }
		check((Decimal(4) / Decimal(2)) == Decimal(2)) { "4 / 2 != 2" }
		check((Decimal(6) / Decimal(2)) == Decimal(3)) { "6 / 2 != 2" }
		check((Decimal(1) / Decimal(0, 0.25)) == Decimal(4)) { "1 / 0.25 != 4" }
	}

	@Test
	@DisplayName("ImpreciseFraction store/load")
	fun storeLoad() {
		run {
			val f = Decimal(4, 0.28)
			val loaded = Decimal.fromByteArray(f.toByteArray())
			check(f == loaded) { "$f != $loaded" }
		}

		run {
			val f = Decimal(32748293658335L, 0.3472302174)
			val loaded = Decimal.fromByteArray(f.toByteArray())
			check(f == loaded) { "$f != $loaded" }
		}
	}

	@Test
	@DisplayName("ImpreciseFraction.valueOf")
	fun valueOf() {
		for (i in -2100 .. 2100) {
			assertEquals(Decimal(i), Decimal.valueOf(i))
		}

		for (i in -2100 .. 2100) {
			assertEquals(Decimal(i.toLong()), Decimal.valueOf(i.toLong()))
		}
	}

	@Test
	@DisplayName("ImpreciseFraction serialization")
	fun serialization() {
		val output = FastByteArrayOutputStream()
		val outputStream = ObjectOutputStream(output)

		val val1 = Decimal("283")
		val val2 = Decimal("0")
		val val3 = Decimal("-4")
		val val4 = Decimal("25.225")

		outputStream.writeObject(val1)
		outputStream.writeObject(val2)
		outputStream.writeObject(val3)
		outputStream.writeObject(val4)

		val input = FastByteArrayInputStream(output.array, 0, output.length)
		val inputStream = ObjectInputStream(input)

		assertEquals(val1, inputStream.readObject())
		assertEquals(val2, inputStream.readObject())
		assertEquals(val3, inputStream.readObject())
		assertEquals(val4, inputStream.readObject())
	}
}
